//+------------------------------------------------------------------+
//|                                           NearestNeighbor_v1.mq4 |
//|   Prediction of future based on the nearest neighbor in the past |
//|                                             Copyright 2010, gpwr |
//|                                               vlad1004@yahoo.com |
//+------------------------------------------------------------------+
#property copyright "gpwr"


#property indicator_chart_window
#property indicator_buffers 2

//--- future model outputs
#property indicator_color1  Red
#property indicator_style1  STYLE_SOLID
#property indicator_width1  1
//--- past model outputs
#property indicator_color2  Blue
#property indicator_style2  STYLE_SOLID
#property indicator_width2  1

//--- indicator inputs
extern int    Npast   =300; // # of past bars in a pattern
extern int    Nfut    =50;  // # of future bars in a pattern (must be < Npast)

//--- indicator buffers
double ynn[],xnn[];

//--- global variables
int    PrevBars,si;
double mx[],sxx[],denx[],sumx,sumxx;
bool   FirstTime;


//+------------------------------------------------------------------+
//| Custom indicator initialization function                         |
//+------------------------------------------------------------------+
void init()
{
//--- indicator buffers mapping
   SetIndexBuffer(0,ynn);
   SetIndexBuffer(1,xnn);
   IndicatorDigits(Digits);
   IndicatorShortName("1NN("+Npast+")");
   SetIndexShift(0, Nfut);
//--- initialize global variables
   PrevBars  = Bars-1;
   FirstTime = true;
   si = Npast*(Npast + 1)/2;
   
   return(0);  
}
//+------------------------------------------------------------------+
//| Custom indicator iteration function                              |
//+------------------------------------------------------------------+
int start()
{
   int limit, counted_bars = IndicatorCounted(); 
   
   if(counted_bars < 0) return(0);
   if(counted_bars < 1)
   { 
   limit = Bars-1;   
      for(int i=limit;i>=0;i--) 
      {
      xnn[i]    = EMPTY_VALUE; 
      ynn[i]    = EMPTY_VALUE; 
      }
   }   

//--- check for insufficient data and new bar
   int bars=Bars;
   if(bars<Npast+Nfut)
   {
   Print("Error: not enough bars in history!");
   return(0);
   }
   if(PrevBars==bars) return(0);
   PrevBars=bars;


//--- main cycle
//--- Compute correlation sums for current pattern
//--- Current pattern starts at i=bars-Npast and ends at i=bars-1
   double my  = 0.0;
   double syy = 0.0;
   for(i=0;i<Npast;i++)
   {
   double y=Open[Bars-1-(bars-Npast+i)];
   my  += y*(i + 1);
   syy += y*y*(i + 1);
   }
   double deny=syy*si-my*my;
   
   if(deny<=0)
   {
   Print("Zero or negative syy*si-my*my = ",deny);
   return(0);
   }
   
   deny=MathSqrt(deny);

//--- Compute correlation sums for past patterns
//--- Past patterns start at k=0 and end at k=bars-Npast-Nfut
   ArrayResize(mx  ,bars-Npast-Nfut+1);
   ArrayResize(sxx ,bars-Npast-Nfut+1);
   ArrayResize(denx,bars-Npast-Nfut+1);
   
   int kstart;
   
   if(FirstTime) kstart=0; else kstart=bars-Npast-Nfut;
   FirstTime=false;
   
   for(int k=kstart;k<=bars-Npast-Nfut;k++)
   {
      if(k == 0)
      {
      mx[0] = 0.0;
      sxx[0]= 0.0;
      sumx  = 0.0;
      sumxx = 0.0;
         for(i=0;i<Npast;i++)
         {
         double x = Open[Bars-1-i];
         double xx= x*x;
         mx[0] += x*(i+1);
         sxx[0]+= xx*(i+1);
         sumx  += x;
         sumxx += xx;
         }
      }
      else
      {
      double xnew = Open[Bars-1-(k+Npast-1)];
      double xold = Open[Bars-1-(k-1)];
      //if(k==bars-Npast-Nfut) Print("inew=",Bars-1-(k+Npast-1)," xnew=",DoubleToStr(xnew,Digits)," iold=",Bars-1-(k+1)," xold=",DoubleToStr(xold,Digits));
      mx[k]  = mx[k-1] - sumx  + xnew*Npast;
      sxx[k] = sxx[k-1]- sumxx + xnew*xnew*Npast;
         
      sumx +=xnew-xold;
      sumxx+=xnew*xnew-xold*xold;
      }
   denx[k]=sxx[k]*si-mx[k]*mx[k];
   }

//--- Compute cross-correlation sums and correlation coefficients and find NN
   double sxy[];
   ArrayResize(sxy,bars-Npast-Nfut+1);
   double b,corrMax=0;
   int knn=0;
   for(k=0;k<=bars-Npast-Nfut;k++)
   {
   //--- Compute sxy
   sxy[k]=0.0;
      for(i=0;i<Npast;i++) sxy[k]+=Open[Bars-1-(k+i)]*Open[Bars-1-(bars-Npast+i)]*(i+1);

   //--- Compute corr coefficient
      
      if(denx[k]<=0)
      {
      Print("Zero or negative sxx[k]*si-mx[k]*mx[k]. Skipping pattern # ",k);
      continue;
      }
       
   double num=sxy[k]*si-mx[k]*my;
   double corr=num/MathSqrt(denx[k])/deny;
   
      if(corr>corrMax)
      {
      corrMax=corr;
      knn=k;
      b=num/denx[k];
      }
   }
   
   Print("Nearest neighbor starts on ",TimeToStr(Time[Bars-1-knn])," and ends on ",TimeToStr(Time[Bars-1-(knn+Npast-1)]),
         ". Its correlation coefficient with current pattern is ",corrMax);

//--- Compute xnn[] and ynn[] by scaling the nearest neighbor
   double delta=Open[0]-b*Open[Bars-1-(knn+Npast-1)];
   
   for(i=0;i<Npast+Nfut;i++)
   {
   if(i<=Npast-1) xnn[Npast-i-1]      = b*Open[Bars-1-(knn+i)]+delta;
   if(i>=Npast-1) ynn[Npast+Nfut-i-1] = b*Open[Bars-1-(knn+i)]+delta;
   }

   return(0);
}
//+------------------------------------------------------------------+

